import json

import pytest
from django.test import Client
import requests_mock


@pytest.mark.django_db
@pytest.mark.parametrize('enable', (False, True))
@pytest.mark.parametrize('as_admin', (False, True))
@pytest.mark.parametrize('posting', (False, True))
def test_cavalry(settings, as_admin, enable, posting, django_user_model):
    settings.CAVALRY_ENABLED = enable
    settings.CAVALRY_ELASTICSEARCH_URL_TEMPLATE = ('http://localhost:59595/asdf/foo' if posting else None)
    client = Client()
    if as_admin:
        user = django_user_model.objects.create_superuser(
            username='adminator',
            email='adminator@example.com',
            password='henlo',
        )
        client.force_login(user)
    else:
        user = None
    with requests_mock.mock() as m:
        m.post(settings.CAVALRY_ELASTICSEARCH_URL_TEMPLATE, status_code=201, json={'ok': True})
        content = client.get('/').content

    # Check precondition: the user seemed logged in
    assert (b'adminator' in content) == as_admin

    # Check that the injection div only appears for admins
    assert (b'<div' in content) == (enable and as_admin)

    # Check that stats are posted only when posting is enabled
    assert bool(m.called) == (enable and posting)
    if enable and posting:
        payload = json.loads(m.request_history[-1].body.decode('utf-8'))
        # Check that user IDs are sent
        assert payload.get('user_id') == (user.id if user else None)



import json

from django.conf import settings
from django.utils.crypto import get_random_string


STYLE = '''
#cv {
position:fixed;
z-index:9999;
bottom:0;
left:50%;
transform: translateX(-50%);
text-align:center;
font:10px/1.1 menlo,consolas,monospace;
background:rgba(0,0,0,0.8);
color:#fff;
padding:1px;
}
'''.replace('\n', '').strip()


def can_inject_stats(request):
    # When debugging, stats can always be shown
    if settings.DEBUG:
        return True

    # When the user (if there is one) is a superuser, stats can be shown
    user = getattr(request, 'user', None)
    if user and getattr(user, 'is_superuser', False):
        return True

    return False


def inject_html(request, response, data, summary_data):
    try:
        body_closing_index = response.content.rindex(b'</body>')
    except ValueError:
        return
    content = ' | '.join(
        '%s=%s' % (key, round(value, 3) if isinstance(value, float) else value)
        for (key, value)
        in sorted(summary_data.items())
    )
    ns = 'cv_%s' % get_random_string()
    html = """<style>{style}</style><div id="{ns}">{content}</div>""".format(
        ns=ns,
        content=content,
        style=STYLE.replace('#cv', '#' + ns),
    )
    script = generate_console_script(data, with_stacks=bool(request.GET.get('_cavalry_stacks')))
    html += "<script>{script}</script>".format(script='\n'.join(script))

    response.content = (
        response.content[:body_closing_index] +
        html.encode('utf-8') +
        response.content[body_closing_index:]
    )

    if 'content-length' in response:
        response['content-length'] = len(response.content)


def generate_console_script(data, with_stacks=False):
    script = []
    for db, data in data.get('databases', {}).items():
        queries = data['queries']
        header = '{db}: {n} queries ({t} msec)'.format(
            db=db,
            n=data['n_queries'],
            t=data['time'],
        )
        script.append('console.group({});'.format(json.dumps(header)))
        for query in queries:
            stack = (query.get('stack', None) if with_stacks else ())
            script.append('console.log({time:.3f}, {sql});'.format(
                time=query.get('hrtime', 0) * 1000, sql=json.dumps(query['sql'])))
            if stack:
                script.append('console.groupCollapsed("Stack");')
                stack_lines = stack.as_lines()
                script.append('console.log({});'.format(json.dumps('\n'.join(stack_lines))))
                script.append('console.groupEnd();')

        script.append('console.groupEnd();')
    return script


def inject_stats(request, response, data):
    if not can_inject_stats(request):
        return
    summary_data = summarize_data(data)
    content_type = response.get('content-type', '').lower()
    if content_type.startswith('text/html') and ('charset' not in content_type or 'utf-8' in content_type):
        inject_html(request, response, data, summary_data)
    response['x-cavalry-data'] = json.dumps(summary_data)


def summarize_data(data):
    summary_data = {
        'request time': data['duration'] * 1000,
    }
    for db, db_data in data.get('databases', {}).items():
        summary_data['queries [%s]' % db] = db_data['n_queries']
        summary_data['time [%s]' % db] = db_data['time']
    return summary_data



import traceback
from contextlib import contextmanager

from django.db import connections
from django.db.backends import utils as db_backend_utils
from django.db.backends.utils import CursorWrapper, logger

from cavalry.locals import get_storage
from cavalry.stack import Stack
from cavalry.timing import get_time


class CavalryCursorDebugWrapper(CursorWrapper):
    # This is an enhanced version of `django.db.backends.utils.CursorDebugWrapper`.

    def execute(self, sql, params=None):
        start = get_time()
        try:
            return super(CavalryCursorDebugWrapper, self).execute(sql, params)
        finally:
            stop = get_time()
            duration = stop - start
            sql = self.db.ops.last_executed_query(self.cursor, sql, params)
            self._record(sql, params, duration)

    def executemany(self, sql, param_list):
        start = get_time()
        try:
            return super(CavalryCursorDebugWrapper, self).executemany(sql, param_list)
        finally:
            stop = get_time()
            duration = stop - start
            try:
                times = len(param_list)
            except TypeError:  # param_list could be an iterator
                times = '?'
            self._record(sql, param_list, duration, times)

    def _record(self, sql, params, duration, times=None):
        if get_storage().get('db_record_stacks'):
            stack = Stack(traceback.extract_stack()[:-2])  # the two last frames are in cavalry
        else:
            stack = []
        self.db.queries_log.append({
            'sql': ('%s times: %s' % (times, sql) if times else sql),
            'time': "%.3f" % duration,
            'hrtime': duration,
            'stack': stack,
        })
        logger.debug(
            '(%.3f) %s; args=%s', duration, sql, params,
            extra={'duration': duration, 'sql': sql, 'params': params}
        )


@contextmanager
def force_debug_cursor():
    for conn in connections.all():
        conn._cavalry_old_force_debug_cursor = conn.force_debug_cursor
        conn.force_debug_cursor = True
    yield
    for conn in connections.all():
        conn.force_debug_cursor = conn._cavalry_old_force_debug_cursor


def patch_db():
    db_backend_utils.CursorDebugWrapper = CavalryCursorDebugWrapper



import json
import platform
from datetime import datetime
from logging import getLogger

import requests
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.encoding import force_bytes

from cavalry.stack import Stack

try:
    from ipware import get_ip
except ImportError:
    def get_ip(request): return request.META.get('REMOTE_ADDR')

log = getLogger(__name__)

sess = requests.Session()  # Persistent requests session, yes


class PayloadJSONEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, Stack):
            if not getattr(settings, 'CAVALRY_POST_STACKS', True):
                return ''
            return '\n'.join(obj.as_lines())
        return super().default(obj)


def post_stats(request, response, data):
    es_url_template = getattr(settings, 'CAVALRY_ELASTICSEARCH_URL_TEMPLATE', None)
    if not es_url_template:
        return
    payload = build_payload(data, request, response)
    es_url = es_url_template.format_map(
        dict(
            payload,
            ymd=datetime.utcnow().strftime('%Y-%m-%d'),
        ),
    )
    body = force_bytes(json.dumps(payload, cls=PayloadJSONEncoder))
    try:
        resp = sess.post(es_url, data=body, headers={'Content-Type': 'application/json'}, timeout=0.5)
        if resp.status_code != 201:
            log.warning('Unable to post data to %s (error %s): %s', es_url, resp.status_code, resp.text)
    except Exception as e:
        log.warning('Unable to post data to %s: %s', es_url, e)


def build_payload(data, request, response):
    payload = {
        'node': platform.node(),
        'content-type': response.get('content-type'),
        'ip': get_ip(request),
        'method': request.method,
        'params': dict(request.GET),
        'path': request.path,
        'status': response.status_code,
        'time': datetime.utcnow(),
    }
    resolver_match = getattr(request, 'resolver_match', None)
    if resolver_match:
        payload.update({
            'view': resolver_match.view_name,
        })
    user = getattr(request, 'user', None)
    if user:
        payload.update({
            'user_id': user.id,
        })
    payload.update(data)
    payload.pop('start_time', None)
    payload.pop('end_time', None)
    payload.pop('db_record_stacks', None)
    return payload



from random import random
from threading import Thread

from django.conf import settings
from django.db import connections
from django.template.response import SimpleTemplateResponse

from cavalry.db import force_debug_cursor, patch_db
from cavalry.locals import managed
from cavalry.poster import post_stats
from cavalry.reporter import inject_stats
from cavalry.timing import get_time


def _process(request, get_response):
    with force_debug_cursor(), managed(
        db_record_stacks=getattr(settings, 'CAVALRY_DB_RECORD_STACKS', True),
    ) as data:
        data['start_time'] = get_time()
        response = get_response(request)
        if isinstance(response, SimpleTemplateResponse):
            response.render()
        data['end_time'] = get_time()
        data['duration'] = data['end_time'] - data['start_time']
        data['databases'] = {}
        for conn in connections.all():
            queries = conn.queries
            data['databases'][conn.alias] = {
                'queries': queries,
                'n_queries': len(queries),
                'time': (sum(q.get('hrtime', 0) * 1000 for q in queries) if queries else 0),
            }
    inject_stats(request, response, data)
    post_stats_kwargs = {'request': request, 'response': response, 'data': data}

    if getattr(settings, 'CAVALRY_THREADED_POST', False):
        Thread(name='cavalry poster', target=post_stats, kwargs=post_stats_kwargs, daemon=False).start()
    else:
        post_stats(**post_stats_kwargs)
    return response


def should_cavalrize(request):
    probability = getattr(settings, 'CAVALRY_PROBABILITY', 1.0)
    return probability >= 1 or (random() < probability)


def cavalry(get_response):
    patch_db()

    def middleware(request):
        if not (getattr(settings, 'CAVALRY_ENABLED', False) and should_cavalrize(request)):
            return get_response(request)
        return _process(request, get_response)

    return middleware
